@@ -214,7 +214,6 @@ class Term(aclgenerator.Term):
 
     Returns:
       list of statements related to ports and protocols.
-
     """
 
     def PortStatement(protocol, source, destination):
@@ -249,14 +248,6 @@ class Term(aclgenerator.Term):
     # TODO: Call self.NormalizeIcmpTypes.
     icmp_type = self.MapICMPtypes(address_family, icmp_type)
 
-    if address_family == mixed:
-      # The way we handle mixed is we call ourselves twice.
-      ipv4_list = self.PortsAndProtocols(ip4, protocol, src_ports, dst_ports,
-                                         icmp_type)
-      ipv6_list = self.PortsAndProtocols(ip6, protocol, src_ports, dst_ports,
-                                         icmp_type)
-      return ipv4_list + ipv6_list
-
     if address_family == 'ip':
       # IPv4 stuff.
       if icmp_type and ('icmp' in ip_protocol):
@@ -334,7 +325,7 @@ class Term(aclgenerator.Term):
     # Base chain already allows all return traffic of
     # state (ESTABLISHED, RELATED)
     # This should prevent invalid, untracked packets from being accepted.
-    if 'deny' not in term.action:
+    if 'deny' not in term.action and not term.icmp_type:
       options.append('ct state new')
 
     # 'logging' handling.
@@ -417,12 +408,12 @@ class Term(aclgenerator.Term):
 
     if src_addr and dst_addr:
       # Condition where term has both defined.
-      if address_family == 'inet' or address_family == 'ip':
+      if address_family == 'ip':
         if src_addr_book['ip'] and dst_addr_book['ip']:
           address_statement.append(
               'ip saddr ' + self.CreateAnonymousSet(src_addr_book['ip']) + ' ' +
               'ip daddr ' + self.CreateAnonymousSet(dst_addr_book['ip']))
-      if address_family == 'inet' or address_family == 'ip6':
+      if address_family == 'ip6':
         if src_addr_book['ip6'] and dst_addr_book['ip6']:
           address_statement.append(
               'ip6 saddr ' + self.CreateAnonymousSet(src_addr_book['ip6']) +
@@ -430,20 +421,20 @@ class Term(aclgenerator.Term):
               self.CreateAnonymousSet(dst_addr_book['ip6']))
     elif src_addr:
       # Term has only src defined.
-      if address_family == 'inet' or address_family == 'ip':
+      if address_family == 'ip':
         if src_addr_book['ip']:
           address_statement.append('ip saddr ' +
                                    self.CreateAnonymousSet(src_addr_book['ip']))
-      if address_family == 'inet' or address_family == 'ip6':
+      if address_family == 'ip6':
         if src_addr_book['ip6']:
           address_statement.append(
               'ip6 saddr ' + self.CreateAnonymousSet(src_addr_book['ip6']))
     elif dst_addr:
-      if address_family == 'inet' or address_family == 'ip':
+      if address_family == 'ip':
         if dst_addr_book['ip']:
           address_statement.append('ip daddr ' +
                                    self.CreateAnonymousSet(dst_addr_book['ip']))
-      if address_family == 'inet' or address_family == 'ip6':
+      if address_family == 'ip6':
         if dst_addr_book['ip6']:
           address_statement.append(
               'ip6 daddr ' + self.CreateAnonymousSet(dst_addr_book['ip6']))
@@ -462,36 +453,38 @@ class Term(aclgenerator.Term):
 
     Returns:
       list of strings. Representing a ruleset for later formatting.
-
     """
     term_ruleset = []
 
-    address_list = []
     # COMMENT handling.
     if self.verbose:
-      for line in self.term.comment:
-        term_ruleset.append('comment "%s"' % line)
-
-    # ADDRESS handling.
-    address_list = self._AddrStatement(self.address_family,
-                                       self.term.source_address,
-                                       self.term.destination_address)
-
-    # PORTS and PROTOCOLS handling.
-    proto_and_ports = self.PortsAndProtocols(self.address_family,
-                                             self.term.protocol,
-                                             self.term.source_port,
-                                             self.term.destination_port,
-                                             self.term.icmp_type)
-
+      term_ruleset.append(
+          'comment ' + aclgenerator.TruncateWords(
+              self.term.comment, Nftables.COMMENT_CHAR_LIMIT))
     # OPTIONS / LOGGING / COUNTERS
     opt = self._OptionsHandler(term)
     # STATEMENT VERDICT / ACTION.
     verdict = self._ACTIONS[self.term.action[0]]
-    # TODO: If verdict is not supported, drop nftable_rule for it.
-    nftable_rule = self.GroupExpressions(address_list, proto_and_ports, opt,
-                                         verdict)
-    term_ruleset.extend(nftable_rule)
+
+    address_families = [self.address_family
+                       ] if self.address_family != mixed else [ip4, ip6]
+    for address_family in address_families:
+      # ADDRESS handling.
+      address_list = self._AddrStatement(address_family,
+                                         self.term.source_address,
+                                         self.term.destination_address)
+
+      # PORTS and PROTOCOLS handling.
+      proto_and_ports = self.PortsAndProtocols(address_family,
+                                               self.term.protocol,
+                                               self.term.source_port,
+                                               self.term.destination_port,
+                                               self.term.icmp_type)
+
+      # TODO: If verdict is not supported, drop nftable_rule for it.
+      nftable_rule = self.GroupExpressions(address_list, proto_and_ports, opt,
+                                           verdict)
+      term_ruleset.extend(nftable_rule)
     return term_ruleset
 
   def _AddressClassifier(self, address_to_classify):
@@ -571,6 +564,8 @@ class Nftables(aclgenerator.ACLGenerator):
       'expiration',
   ])
 
+  COMMENT_CHAR_LIMIT = 126
+
   _AF_MAP = {'inet': (4,), 'inet6': (6,), 'mixed': (4, 6)}
   # Below mapping converts capirca HEADER native to nftables table.
   # In Nftables 'inet' contains both IPv4 and IPv6 addresses and rules.
@@ -796,9 +791,16 @@ class Nftables(aclgenerator.ACLGenerator):
         # base chain header and contents.
         nft_config.append(TabSpacer(4, 'chain %s {' % item))
         if base_chain_dict[item]['comment']:
-          # Handle multi-line comments
-          for comment in base_chain_dict[item]['comment']:
-            nft_config.append(TabSpacer(8, 'comment "%s"' % comment))
+          # Due to Nftables limits on comments, we handle this twice.
+          # First time we comment it out so .nft file is human-readable.
+          nft_config.append(
+              TabSpacer(8, '#' + ' '.join(base_chain_dict[item]['comment'])))
+          # Second time so 'nft list ruleset' keeps the comment in memory.
+          nft_config.append(
+              TabSpacer(
+                  8, 'comment ' +
+                  aclgenerator.TruncateWords(
+              base_chain_dict[item]['comment'], self.COMMENT_CHAR_LIMIT)))
         nft_config.append(
             TabSpacer(
                 8, 'type filter hook %s priority %s; policy %s;' %
