@@ -319,6 +319,7 @@ class Term:
     dscp-except: VarType.DSCP_EXCEPT
     comments: VarType.COMMENT
     encapsulate: VarType.ENCAPSULATE
+    decapsulate: VarType.DECAPSULATE
     filter-term: VarType.FILTER_TERM
     flexible-match-range: VarType.FLEXIBLE_MATCH_RANGE
     forwarding-class: VarType.FORWARDING_CLASS
@@ -333,6 +334,10 @@ class Term:
     pan-application: VarType.PAN_APPLICATION
     policer: VarType.POLICER
     priority: VarType.PRIORITY
+    destination-zone: VarType.DZONE
+    source-service-accounts: VarType.SOURCE_SERVICE_ACCOUNTS
+    target-service-accounts: VarType.TARGET_SERVICE_ACCOUNTS
+    source-zone: VarType.SZONE
     vpn: VarType.VPN
   """
   ICMP_TYPE = {4: {'echo-reply': 0,
@@ -456,20 +461,25 @@ class Term:
     self.destination_prefix_except = []
     self.inactive = False
     self.encapsulate = None
+    self.decapsulate = None
     self.port_mirror = None
     # srx specific
+    self.destination_zone = []
+    self.source_zone = []
     self.vpn = None
     # gce specific
     self.source_tag = []
     self.destination_tag = []
     self.priority = None
+    self.source_service_accounts = []
+    self.target_service_accounts = []
     # iptables specific
     self.source_interface = None
     self.destination_interface = None
     self.platform = []
     self.platform_exclude = []
     self.target_resources = []
-    self.target_service_accounts = []
+
     self.timeout = None
     self.flattened = False
     self.flattened_addr = None
@@ -627,6 +637,9 @@ class Term:
     if self.encapsulate:
       if not other.encapsulate:
         return False
+    if self.decapsulate:
+      if not other.decapsulate:
+        return False
     if self.fragment_offset:
       # fragment_offset looks like 'integer-integer' or just, 'integer'
       sfo = sorted([int(x) for x in self.fragment_offset.split('-')])
@@ -677,6 +690,14 @@ class Term:
       if sorted(self.platform_exclude) is not sorted(other.platform_exclude):
         return False
 
+    if self.source_zone:
+      if sorted(self.source_zone) is not sorted(other.source_zone):
+        return False
+
+    if self.destination_zone:
+      if sorted(self.destination_zone) is not sorted(other.destination_zone):
+        return False
+
     # we have containment
     return True
 
@@ -705,6 +726,9 @@ class Term:
       ret_str.append('  destination_tag: %s' % self.destination_tag)
     if self.target_resources:
       ret_str.append('  target_resources: %s' % self.target_resources)
+    if self.source_service_accounts:
+      ret_str.append('  source_service_accounts: %s' %
+                     self.source_service_accounts)
     if self.target_service_accounts:
       ret_str.append('  target_service_accounts: %s' %
                      self.target_service_accounts)
@@ -732,6 +756,8 @@ class Term:
       ret_str.append('  next_ip: %s' % self.next_ip)
     if self.encapsulate:
       ret_str.append('  encapsulate: %s' % self.encapsulate)
+    if self.decapsulate:
+      ret_str.append('  decapsulate: %s' % self.decapsulate)
     if self.protocol:
       ret_str.append('  protocol: %s' % sorted(self.protocol))
     if self.protocol_except:
@@ -790,6 +816,10 @@ class Term:
                        (vpn_name, pair_policy))
       else:
         ret_str.append('  vpn: name = %s' % vpn_name)
+    if self.source_zone:
+      ret_str.append('  source_zone: %s' % sorted(self.source_zone))
+    if self.destination_zone:
+      ret_str.append('  destination_zone: %s' % sorted(self.destination_zone))
 
     return '\n'.join(ret_str)
 
@@ -929,6 +959,10 @@ class Term:
     if self.encapsulate != other.encapsulate:
       return False
 
+    # decapsulate
+    if self.decapsulate != other.decapsulate:
+      return False
+
     # flexible_match
     if self.flexible_match_range != other.flexible_match_range:
       return False
@@ -937,6 +971,14 @@ class Term:
     if self.port_mirror != other.port_mirror:
       return False
 
+    # source_zone
+    if sorted(self.source_zone) != sorted(other.source_zone):
+      return False
+
+    # destination_zone
+    if sorted(self.destination_zone) != sorted(other.destination_zone):
+      return False
+
     return True
 
   def __ne__(self, other):
@@ -987,11 +1029,15 @@ class Term:
         counter += self._IPV4_BYTE_SIZE
     return counter
 
-  def FlattenAll(self):
+  def FlattenAll(self, mutate=True):
     """Reduce source, dest, and address fields to their post-exclude state.
 
     Populates the self.flattened_addr, self.flattened_saddr,
     self.flattened_daddr by removing excludes from includes.
+
+    Args:
+      mutate: Boolean value indicating if this method should mutate the original
+              address (address, destination_address, source_address)
     """
     # No excludes, set flattened attributes and move along.
     self.flattened = True
@@ -1007,17 +1053,20 @@ class Term:
           self.source_address,
           self.source_address_exclude,
           collapse_addrs=False)
-      self.source_address = self.flattened_saddr
+      if mutate:
+        self.source_address = self.flattened_saddr
     if self.destination_address_exclude:
       self.flattened_daddr = nacaddr.AddressListExclude(
           self.destination_address,
           self.destination_address_exclude,
           collapse_addrs=False)
-      self.destination_address = self.flattened_daddr
+      if mutate:
+        self.destination_address = self.flattened_daddr
     if self.address_exclude:
       self.flattened_addr = nacaddr.AddressListExclude(
           self.address, self.address_exclude, collapse_addrs=False)
-      self.address = self.flattened_addr
+      if mutate:
+        self.address = self.flattened_addr
 
   def GetAddressOfVersion(self, addr_type, af=None):
     """Returns addresses of the appropriate Address Family.
@@ -1128,8 +1177,14 @@ class Term:
           self.flexible_match_range.append(x.value)
         elif x.var_type is VarType.TARGET_RESOURCES:
           self.target_resources.append(x.value)
+        elif x.var_type is VarType.SOURCE_SERVICE_ACCOUNTS:
+          self.source_service_accounts.append(x.value)
         elif x.var_type is VarType.TARGET_SERVICE_ACCOUNTS:
           self.target_service_accounts.append(x.value)
+        elif x.var_type is VarType.SZONE:
+          self.source_zone.append(x.value)
+        elif x.var_type is VarType.DZONE:
+          self.destination_zone.append(x.value)
         else:
           raise TermObjectTypeError(
               '%s isn\'t a type I know how to deal with (contains \'%s\')' % (
@@ -1168,6 +1223,8 @@ class Term:
         self.counter = obj
       elif obj.var_type is VarType.ENCAPSULATE:
         self.encapsulate = obj.value
+      elif obj.var_type is VarType.DECAPSULATE:
+        self.decapsulate = obj.value
       elif obj.var_type is VarType.PORT_MIRROR:
         self.port_mirror = obj.value
       elif obj.var_type is VarType.TRAFFIC_CLASS_COUNT:
@@ -1213,6 +1270,8 @@ class Term:
         self.ttl = int(obj.value)
       elif obj.var_type is VarType.TARGET_RESOURCES:
         self.target_resources.append(obj.value)
+      elif obj.var_type is VarType.SOURCE_SERVICE_ACCOUNTS:
+        self.source_service_accounts.append(obj.value)
       elif obj.var_type is VarType.TARGET_SERVICE_ACCOUNTS:
         self.target_service_accounts.append(obj.value)
       elif obj.var_type is VarType.FILTER_TERM:
@@ -1251,7 +1310,7 @@ class Term:
             'term "%s" has both verbatim and non-verbatim tokens.' % self.name)
     else:
       if (not self.action and not self.routing_instance and not self.next_ip and
-          not self.encapsulate and not self.filter_term and
+          not self.encapsulate and not self.decapsulate and not self.filter_term and
           not self.port_mirror):
         raise TermNoActionError('no action specified for term %s' % self.name)
       if self.filter_term and self.action:
@@ -1530,6 +1589,10 @@ class VarType:
   FILTER_TERM = 62
   RESTRICT_ADDRESS_FAMILY = 63
   PORT_MIRROR = 64
+  SZONE = 65
+  DZONE = 66
+  DECAPSULATE = 67
+  SOURCE_SERVICE_ACCOUNTS = 68
 
 
   def __init__(self, var_type, value):
@@ -1693,6 +1756,7 @@ tokens = (
     'COUNTER',
     'DADDR',
     'DADDREXCLUDE',
+    'DECAPSULATE',
     'DINTERFACE',
     'DPFX',
     'EDPFX',
@@ -1704,6 +1768,7 @@ tokens = (
     'DSCP_RANGE',
     'DSCP_SET',
     'DTAG',
+    'DZONE',
     'ENCAPSULATE',
     'ESCAPEDSTRING',
     'ETHER_TYPE',
@@ -1748,9 +1813,11 @@ tokens = (
     'SADDR',
     'SADDREXCLUDE',
     'SINTERFACE',
+    'SOURCE_SERVICE_ACCOUNTS',
     'SPFX',
     'ESPFX',
     'SPORT',
+    'SZONE',
     'STAG',
     'STRING',
     'TARGET',
@@ -1779,6 +1846,7 @@ reserved = {
     'restrict-address-family': 'RESTRICT_ADDRESS_FAMILY',
     'comment': 'COMMENT',
     'counter': 'COUNTER',
+    'decapsulate': 'DECAPSULATE',
     'destination-address': 'DADDR',
     'destination-exclude': 'DADDREXCLUDE',
     'destination-interface': 'DINTERFACE',
@@ -1786,6 +1854,7 @@ reserved = {
     'destination-prefix-except': 'EDPFX',
     'destination-port': 'DPORT',
     'destination-tag': 'DTAG',
+    'destination-zone': 'DZONE',
     'dscp-except': 'DSCP_EXCEPT',
     'dscp-match': 'DSCP_MATCH',
     'dscp-set': 'DSCP_SET',
@@ -1827,10 +1896,12 @@ reserved = {
     'source-address': 'SADDR',
     'source-exclude': 'SADDREXCLUDE',
     'source-interface': 'SINTERFACE',
+    'source-service-accounts': 'SOURCE_SERVICE_ACCOUNTS',
     'source-prefix': 'SPFX',
     'source-prefix-except': 'ESPFX',
     'source-port': 'SPORT',
     'source-tag': 'STAG',
+    'source-zone': 'SZONE',
     'target': 'TARGET',
     'target-resources': 'TARGET_RESOURCES',
     'target-service-accounts': 'TARGET_SERVICE_ACCOUNTS',
@@ -1972,6 +2043,7 @@ def p_term_spec(p):
                 | term_spec dscp_set_spec
                 | term_spec dscp_match_spec
                 | term_spec dscp_except_spec
+                | term_spec decapsulate_spec
                 | term_spec encapsulate_spec
                 | term_spec ether_type_spec
                 | term_spec exclude_spec
@@ -2004,6 +2076,8 @@ def p_term_spec(p):
                 | term_spec qos_spec
                 | term_spec pan_application_spec
                 | term_spec routinginstance_spec
+                | term_spec source_service_accounts_spec
+                | term_spec term_zone_spec
                 | term_spec tag_list_spec
                 | term_spec target_resources_spec
                 | term_spec target_service_accounts_spec
@@ -2105,6 +2179,11 @@ def p_encapsulate_spec(p):
   p[0] = VarType(VarType.ENCAPSULATE, p[4])
 
 
+def p_decapsulate_spec(p):
+  """ decapsulate_spec : DECAPSULATE ':' ':' STRING """
+  p[0] = VarType(VarType.DECAPSULATE, p[4])
+
+
 def p_port_mirror_spec(p):
   """ port_mirror_spec : PORT_MIRROR ':' ':' STRING """
   p[0] = VarType(VarType.PORT_MIRROR, p[4])
@@ -2253,6 +2332,11 @@ def p_protocol_spec(p):
   for proto in p[4]:
     p[0].append(VarType(VarType.PROTOCOL, proto))
 
+def p_source_service_accounts_spec(p):
+  """ source_service_accounts_spec : SOURCE_SERVICE_ACCOUNTS ':' ':' one_or_more_strings """
+  p[0] = []
+  for service_account in p[4]:
+    p[0].append(VarType(VarType.SOURCE_SERVICE_ACCOUNTS, service_account))
 
 def p_tag_list_spec(p):
   """ tag_list_spec : DTAG ':' ':' one_or_more_strings
@@ -2358,6 +2442,15 @@ def p_verbatim_spec(p):
                     | VERBATIM ':' ':' STRING ESCAPEDSTRING """
   p[0] = VarType(VarType.VERBATIM, [p[4], p[5].strip('"').replace('\\"', '"')])
 
+def p_term_zone_spec(p):
+  """ term_zone_spec : SZONE ':' ':' one_or_more_strings
+                     | DZONE ':' ':' one_or_more_strings """
+  p[0] = []
+  for zone in p[4]:
+    if p[1].find('source-zone') >= 0:
+      p[0].append(VarType(VarType.SZONE, zone))
+    elif p[1].find('destination-zone') >= 0:
+      p[0].append(VarType(VarType.DZONE, zone))
 
 def p_vpn_spec(p):
   """ vpn_spec : VPN ':' ':' STRING STRING
